@@ -1,4 +1,5 @@
-# Open Generation and Transmission Operation and Expansion Planning Model with RES and ESS (openTEPES) - February 18, 2021
+""" Open Generation and Transmission Operation and Expansion Planning Model with RES and ESS (openTEPES) - April 7, 2021
+"""
 
 import time
 import math
@@ -6,8 +7,10 @@ import os
 import pandas        as pd
 from   pyomo.environ import DataPortal, Set, Param, Var, Binary, NonNegativeReals, Reals, UnitInterval, Boolean, Any
 
+
 def InputData(DirName, CaseName, mTEPES):
-    print('Input data                  ****')
+    print('Input data                             ****')
+
     _path = os.path.join(DirName, CaseName)
     StartTime = time.time()
     #%% reading data from CSV
@@ -24,6 +27,7 @@ def InputData(DirName, CaseName, mTEPES):
     dfVariableMinStorage = pd.read_csv(_path+'/oT_Data_MinimumStorage_'       +CaseName+'.csv', index_col=[0,1,2])
     dfVariableMaxStorage = pd.read_csv(_path+'/oT_Data_MaximumStorage_'       +CaseName+'.csv', index_col=[0,1,2])
     dfEnergyInflows      = pd.read_csv(_path+'/oT_Data_EnergyInflows_'        +CaseName+'.csv', index_col=[0,1,2])
+    dfEnergyOutflows     = pd.read_csv(_path+'/oT_Data_EnergyOutflows_'       +CaseName+'.csv', index_col=[0,1,2])
     dfNodeLocation       = pd.read_csv(_path+'/oT_Data_NodeLocation_'         +CaseName+'.csv', index_col=[0    ])
     dfNetwork            = pd.read_csv(_path+'/oT_Data_Network_'              +CaseName+'.csv', index_col=[0,1,2])
 
@@ -41,6 +45,7 @@ def InputData(DirName, CaseName, mTEPES):
     dfVariableMinStorage.fillna(0.0, inplace=True)
     dfVariableMaxStorage.fillna(0.0, inplace=True)
     dfEnergyInflows.fillna     (0.0, inplace=True)
+    dfEnergyOutflows.fillna    (0.0, inplace=True)
     dfNodeLocation.fillna      (0.0, inplace=True)
     dfNetwork.fillna           (0.0, inplace=True)
 
@@ -54,6 +59,7 @@ def InputData(DirName, CaseName, mTEPES):
     print('Minimum storage              \n', dfVariableMinStorage.describe())
     print('Maximum storage              \n', dfVariableMaxStorage.describe())
     print('Energy inflows               \n', dfEnergyInflows.describe()     )
+    print('Energy outflows              \n', dfEnergyOutflows.describe()    )
     print('Network                      \n', dfNetwork.describe()           )
 
     #%% reading the sets
@@ -101,6 +107,7 @@ def InputData(DirName, CaseName, mTEPES):
     pIndBinNetInvest     = dfOption   ['IndBinNetInvest'    ][0].astype('int')                                                            # Indicator of binary network    expansion decisions, 0 continuous - 1 binary
     pIndBinGenOperat     = dfOption   ['IndBinGenOperat'    ][0].astype('int')                                                            # Indicator of binary generation operation decisions, 0 continuous - 1 binary
     pIndNetLosses        = dfOption   ['IndNetLosses'       ][0].astype('int')                                                            # Indicator of network losses,                        0 lossless   - 1 ohmic losses
+    pIndBinLineCommit    = dfOption   ['IndBinLineCommit'   ][0].astype('int')                                                            # Indicator of binary switching decisions,            0 continuous - 1 binary
     pENSCost             = dfParameter['ENSCost'            ][0] * 1e-3                                                                   # cost of energy not served           [MEUR/GWh]
     pCO2Cost             = dfParameter['CO2Cost'            ][0]                                                                          # cost of CO2 emission                [EUR/t CO2]
     pUpReserveActivation = dfParameter['UpReserveActivation'][0]                                                                          # upward   reserve activation         [p.u.]
@@ -112,7 +119,7 @@ def InputData(DirName, CaseName, mTEPES):
 
     pScenProb            = dfScenario          ['Probability'  ]                                                                          # probabilities of scenarios          [p.u.]
     pDuration            = dfDuration          ['Duration'     ] * pTimeStep                                                              # duration of load levels             [h]
-    pDemand              = dfDemand            [list(mTEPES.nd)] * 1e-3                                                                   # demand                              [GW]
+    pDemand              = dfDemand            [list(mTEPES.nd)] * 1e-3                                                                   #           demand                    [GW]
     pOperReserveUp       = dfUpOperatingReserve[list(mTEPES.ar)] * 1e-3                                                                   # upward   operating reserve          [GW]
     pOperReserveDw       = dfDwOperatingReserve[list(mTEPES.ar)] * 1e-3                                                                   # downward operating reserve          [GW]
     pVariableMinPower    = dfVariableMinPower  [list(mTEPES.gg)] * 1e-3                                                                   # dynamic variable minimum power      [GW]
@@ -120,6 +127,7 @@ def InputData(DirName, CaseName, mTEPES):
     pVariableMinStorage  = dfVariableMinStorage[list(mTEPES.gg)]                                                                          # dynamic variable minimum storage    [GWh]
     pVariableMaxStorage  = dfVariableMaxStorage[list(mTEPES.gg)]                                                                          # dynamic variable maximum storage    [GWh]
     pEnergyInflows       = dfEnergyInflows     [list(mTEPES.gg)] * 1e-3                                                                   # dynamic energy inflows              [GW]
+    pEnergyOutflows      = dfEnergyOutflows    [list(mTEPES.gg)] * 1e-3                                                                   # dynamic energy outflows             [GW]
 
     # compute the demand as the mean over the time step load levels and assign it to active load levels. Idem for operating reserve, variable max power, variable min and max storage and inflows
     if pTimeStep > 1:
@@ -131,6 +139,7 @@ def InputData(DirName, CaseName, mTEPES):
         pVariableMinStorage = pVariableMinStorage.rolling(pTimeStep).mean()
         pVariableMaxStorage = pVariableMaxStorage.rolling(pTimeStep).mean()
         pEnergyInflows      = pEnergyInflows.rolling     (pTimeStep).mean()
+        pEnergyOutflows     = pEnergyOutflows.rolling    (pTimeStep).mean()
 
     pDemand.fillna            (0.0, inplace=True)
     pOperReserveUp.fillna     (0.0, inplace=True)
@@ -140,6 +149,7 @@ def InputData(DirName, CaseName, mTEPES):
     pVariableMinStorage.fillna(0.0, inplace=True)
     pVariableMaxStorage.fillna(0.0, inplace=True)
     pEnergyInflows.fillna     (0.0, inplace=True)
+    pEnergyOutflows.fillna    (0.0, inplace=True)
 
     if pTimeStep > 1:
         # assign duration 0 to load levels not being considered, active load levels are at the end of every pTimeStep
@@ -169,7 +179,8 @@ def InputData(DirName, CaseName, mTEPES):
     pRatedMinStorage    = dfGeneration  ['MinimumStorage'      ]                                                                            # minimum ESS storage                         [GWh]
     pRatedMaxStorage    = dfGeneration  ['MaximumStorage'      ]                                                                            # maximum ESS storage                         [GWh]
     pEfficiency         = dfGeneration  ['Efficiency'          ]                                                                            #         ESS efficiency                      [p.u.]
-    pStorageType        = dfGeneration  ['StorageType'         ]                                                                            #         ESS type
+    pStorageType        = dfGeneration  ['StorageType'         ]                                                                            #         ESS storage  type
+    pOutflowsType       = dfGeneration  ['OutflowsType'        ]                                                                            #         ESS outflows type
     pRMaxReactivePower  = dfGeneration  ['MaximumReactivePower'] * 1e-3                                                                     # rated maximum reactive power                [Gvar]
 
     pLinearOperCost     = pLinearFuelCost + pCO2EmissionCost
@@ -190,6 +201,9 @@ def InputData(DirName, CaseName, mTEPES):
     pLineNTCBck         = dfNetwork     ['TTCBck'              ] * 1e-3 * dfNetwork['SecurityFactor' ]                                      # net transfer capacity in backward direction [GW]
     pNetFixedCost       = dfNetwork     ['FixedCost'           ] *        dfNetwork['FixedChargeRate']                                      # network    fixed cost                       [MEUR]
     pIndBinLineInvest   = dfNetwork     ['BinaryInvestment'    ]                                                                            # binary line    investment decision          [Yes]
+    pIndBinSwitching    = dfNetwork     ['BinarySwitching'     ]                                                                            # binary line    switching  decision          [Yes]
+    pSwitchOnTime       = dfNetwork     ['SwOnTime'            ]                                                                            # minimum on time                             [h]
+    pSwitchOffTime      = dfNetwork     ['SwOnTime'            ]                                                                            # minimum off time                            [h]
     pAngMin             = dfNetwork     ['AngMin'              ] * math.pi / 180                                                            # Min phase angle difference                  [rad]
     pAngMax             = dfNetwork     ['AngMax'              ] * math.pi / 180                                                            # Max phase angle difference                  [rad]
 
@@ -200,7 +214,7 @@ def InputData(DirName, CaseName, mTEPES):
 
     ReadingDataTime = time.time() - StartTime
     StartTime       = time.time()
-    print('Reading    input data                 ... ', round(ReadingDataTime), 's')
+    print('Reading    input data                  ... ', round(ReadingDataTime), 's')
 
     #%% defining subsets: active load levels (n,n2), thermal units (t), RES units (r), ESS units (es), candidate gen units (gc), candidate ESS units (ec), all the lines (la), candidate lines (lc), candidate DC lines (cd), existing DC lines (cd), lines with losses (ll), reference node (rf), and reactive generating units (gq)
     mTEPES.sc = Set(initialize=mTEPES.scc,                    ordered=True , doc='scenarios'          , filter=lambda mTEPES,scc     :  scc       in mTEPES.scc and pScenProb        [scc] >  0.0)
@@ -243,7 +257,11 @@ def InputData(DirName, CaseName, mTEPES):
 
     # replacing string values by numerical values
     idxDict = dict()
-    idxDict[0] = 0
+    idxDict[0    ] = 0
+    idxDict['No' ] = 0
+    idxDict['NO' ] = 0
+    idxDict['N'  ] = 0
+    idxDict['n'  ] = 0
     idxDict['Yes'] = 1
     idxDict['YES'] = 1
     idxDict['Y'  ] = 1
@@ -251,6 +269,7 @@ def InputData(DirName, CaseName, mTEPES):
 
     pIndBinUnitInvest = pIndBinUnitInvest.map(idxDict)
     pIndBinLineInvest = pIndBinLineInvest.map(idxDict)
+    pIndBinSwitching  = pIndBinSwitching.map (idxDict)
 
     # line type
     pLineType = pLineType.reset_index()
@@ -338,17 +357,31 @@ def InputData(DirName, CaseName, mTEPES):
     pDwTime = round(pDwTime/pTimeStep).astype('int')
 
     #%% definition of the time-steps leap to observe the stored energy at ESS
-    pCycleTimeStep = (pUpTime*0).astype('int')
+    pCycleTimeStep    = (pUpTime*0).astype('int')
+    pOutflowsTimeStep = (pUpTime*0).astype('int')
     for es in mTEPES.es:
-        if  pStorageType[es] == 'Daily'  :
+        if  pStorageType  [es] == 'Daily'  :
             pCycleTimeStep[es] = 1
-        if  pStorageType[es] == 'Weekly' :
+        if  pStorageType  [es] == 'Weekly' :
             pCycleTimeStep[es] = int( 24/pTimeStep)
-        if  pStorageType[es] == 'Monthly':
+        if  pStorageType  [es] == 'Monthly':
             pCycleTimeStep[es] = int(168/pTimeStep)
 
-    pStorageType = pStorageType[pStorageType.index.isin(mTEPES.es)]
-    mTEPES.pStorageType = Param(mTEPES.es, initialize=pStorageType.to_dict(), within=Any)
+        if  pOutflowsType    [es] == 'Hourly' :
+            pOutflowsTimeStep[es] =    1
+        if  pOutflowsType    [es] == 'Daily'  :
+            pOutflowsTimeStep[es] =   24/pTimeStep
+        if  pOutflowsType    [es] == 'Weekly' :
+            pOutflowsTimeStep[es] =  168/pTimeStep
+        if  pOutflowsType    [es] == 'Monthly':
+            pOutflowsTimeStep[es] =  672/pTimeStep
+        if  pOutflowsType    [es] == 'Yearly' :
+            pOutflowsTimeStep[es] = 8736/pTimeStep
+
+        pCycleTimeStep[es] = min(pCycleTimeStep[es], pOutflowsTimeStep[es])
+
+    # the stage duration is the maximum between the defined stage duration and the storage type and the outflows type for any ESS to avoid breaking the energy outflows constraints
+    pStageDuration = max(pStageDuration, pCycleTimeStep.max(), pOutflowsTimeStep.max())
 
     # drop load levels with duration 0
     pDuration         = pDuration.loc        [list(                   mTEPES.n          )]
@@ -359,6 +392,7 @@ def InputData(DirName, CaseName, mTEPES):
     pMaxPower         = pMaxPower.loc        [list(mTEPES.sc*mTEPES.p*mTEPES.n          )]
     pMaxPower2ndBlock = pMaxPower2ndBlock.loc[list(mTEPES.sc*mTEPES.p*mTEPES.n          )]
     pEnergyInflows    = pEnergyInflows.loc   [list(mTEPES.sc*mTEPES.p*mTEPES.n          )]
+    pEnergyOutflows   = pEnergyOutflows.loc  [list(mTEPES.sc*mTEPES.p*mTEPES.n          )]
     pMinStorage       = pMinStorage.loc      [list(mTEPES.sc*mTEPES.p*mTEPES.n          )]
     pMaxStorage       = pMaxStorage.loc      [list(mTEPES.sc*mTEPES.p*mTEPES.n          )]
     pIniInventory     = pIniInventory.loc    [list(mTEPES.sc*mTEPES.p*mTEPES.n          )]
@@ -374,6 +408,7 @@ def InputData(DirName, CaseName, mTEPES):
     pMaxPower2ndBlock[pMaxPower2ndBlock < pEpsilon] = 0.0
     pMaxCharge       [pMaxCharge        < pEpsilon] = 0.0
     pEnergyInflows   [pEnergyInflows    < pEpsilon/pTimeStep] = 0.0
+    pEnergyOutflows  [pEnergyOutflows   < pEpsilon/pTimeStep] = 0.0
     pLineNTCFrw      [pLineNTCFrw       < pEpsilon] = 0.0
     pLineNTCBck      [pLineNTCBck       < pEpsilon] = 0.0
 
@@ -381,7 +416,8 @@ def InputData(DirName, CaseName, mTEPES):
     pMinStorage      [pMinStorage       < pEpsilon] = 0.0
     pMaxStorage      [pMaxStorage       < pEpsilon] = 0.0
     pIniInventory    [pIniInventory     < pEpsilon] = 0.0
-    pInitialInventory[pInitialInventory < pEpsilon] = 0.0
+
+    pInitialInventory = pInitialInventory.where(pInitialInventory > pEpsilon, other= 0.0)
 
     # BigM maximum flow to be used in the Kirchhoff's 2nd law disjunctive constraint
     pBigMFlowBck = pLineNTCBck*0.0
@@ -400,10 +436,11 @@ def InputData(DirName, CaseName, mTEPES):
     # this option avoids a warning in the following assignments
     pd.options.mode.chained_assignment = None
 
-    mTEPES.pIndBinGenInvest      = Param(initialize=pIndBinGenInvest    , within=Boolean, mutable=True)
-    mTEPES.pIndBinNetInvest      = Param(initialize=pIndBinNetInvest    , within=Boolean, mutable=True)
-    mTEPES.pIndBinGenOperat      = Param(initialize=pIndBinGenOperat    , within=Boolean, mutable=True)
-    mTEPES.pIndNetLosses         = Param(initialize=pIndNetLosses       , within=Boolean, mutable=True)
+    mTEPES.pIndBinGenInvest      = Param(initialize=pIndBinGenInvest    , within=Boolean, doc='Indicator of binary generation investment decisions', mutable=True)
+    mTEPES.pIndBinGenOperat      = Param(initialize=pIndBinGenOperat    , within=Boolean, doc='Indicator of binary generation operation  decisions', mutable=True)
+    mTEPES.pIndBinNetInvest      = Param(initialize=pIndBinNetInvest    , within=Boolean, doc='Indicator of binary network    investment decisions', mutable=True)
+    mTEPES.pIndBinLineCommit     = Param(initialize=pIndBinLineCommit   , within=Boolean, doc='Indicator of binary network    switicing  decisions', mutable=True)
+    mTEPES.pIndNetLosses         = Param(initialize=pIndNetLosses       , within=Boolean, doc='Indicator of binary network ohmic losses',            mutable=True)
 
     mTEPES.pENSCost              = Param(initialize=pENSCost            , within=NonNegativeReals)
     mTEPES.pCO2Cost              = Param(initialize=pCO2Cost            , within=NonNegativeReals)
@@ -424,6 +461,7 @@ def InputData(DirName, CaseName, mTEPES):
     mTEPES.pMaxPower             = Param(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.gg, initialize=pMaxPower.stack().to_dict()        , within=NonNegativeReals, doc='Maximum power'                )
     mTEPES.pMaxPower2ndBlock     = Param(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.gg, initialize=pMaxPower2ndBlock.stack().to_dict(), within=NonNegativeReals, doc='Second block'                 )
     mTEPES.pEnergyInflows        = Param(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.gg, initialize=pEnergyInflows.stack().to_dict()   , within=NonNegativeReals, doc='Energy inflows'               )
+    mTEPES.pEnergyOutflows       = Param(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.gg, initialize=pEnergyOutflows.stack().to_dict()  , within=NonNegativeReals, doc='Energy outflows'              )
     mTEPES.pMinStorage           = Param(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.gg, initialize=pMinStorage.stack().to_dict()      , within=NonNegativeReals, doc='ESS Minimum stoarage capacity')
     mTEPES.pMaxStorage           = Param(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.gg, initialize=pMaxStorage.stack().to_dict()      , within=NonNegativeReals, doc='ESS Maximum stoarage capacity')
 
@@ -443,6 +481,7 @@ def InputData(DirName, CaseName, mTEPES):
     mTEPES.pIndBinUnitInvest     = Param(                               mTEPES.gg, initialize=pIndBinUnitInvest.to_dict()        , within=NonNegativeReals, doc='Binary investment decision'   )
     mTEPES.pEfficiency           = Param(                               mTEPES.gg, initialize=pEfficiency.to_dict()              , within=NonNegativeReals, doc='Round-trip efficiency'        )
     mTEPES.pCycleTimeStep        = Param(                               mTEPES.gg, initialize=pCycleTimeStep.to_dict()           , within=NonNegativeReals, doc='ESS Storage cycle'            )
+    mTEPES.pOutflowsTimeStep     = Param(                               mTEPES.gg, initialize=pOutflowsTimeStep.to_dict()        , within=NonNegativeReals, doc='ESS Outflows cycle'           )
     mTEPES.pIniInventory         = Param(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.gg, initialize=pIniInventory.stack().to_dict()    , within=NonNegativeReals, doc='ESS Initial storage',         mutable=True)
 
     mTEPES.pLineLossFactor       = Param(                               mTEPES.ln, initialize=pLineLossFactor.to_dict()          , within=NonNegativeReals, doc='Loss factor'                  )
@@ -456,6 +495,9 @@ def InputData(DirName, CaseName, mTEPES):
     mTEPES.pLineNTCBck           = Param(                               mTEPES.ln, initialize=pLineNTCBck.to_dict()              , within=NonNegativeReals, doc='NTC backward'                 )
     mTEPES.pNetFixedCost         = Param(                               mTEPES.ln, initialize=pNetFixedCost.to_dict()            , within=NonNegativeReals, doc='Network fixed cost'           )
     mTEPES.pIndBinLineInvest     = Param(                               mTEPES.ln, initialize=pIndBinLineInvest.to_dict()        , within=NonNegativeReals, doc='Binary investment decision'   )
+    mTEPES.pSwOnTime             = Param(                               mTEPES.ln, initialize=pSwitchOnTime.to_dict()            , within=NonNegativeReals, doc='Minimum switching on  time'   )
+    mTEPES.pSwOffTime            = Param(                               mTEPES.ln, initialize=pSwitchOffTime.to_dict()           , within=NonNegativeReals, doc='Minimum switching off time'   )
+    mTEPES.pIndBinSwitching      = Param(                               mTEPES.ln, initialize=pIndBinSwitching.to_dict()         , within=NonNegativeReals, doc='Switching decision'           )
     mTEPES.pBigMFlowBck          = Param(                               mTEPES.ln, initialize=pBigMFlowBck.to_dict()             , within=NonNegativeReals, doc='Maximum backward capacity',   mutable=True)
     mTEPES.pBigMFlowFrw          = Param(                               mTEPES.ln, initialize=pBigMFlowFrw.to_dict()             , within=NonNegativeReals, doc='Maximum forward  capacity',   mutable=True)
     mTEPES.pMaxTheta             = Param(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.nd, initialize=pMaxTheta.stack().to_dict()        , within=NonNegativeReals, doc='Maximum voltage angle',       mutable=True)
@@ -472,6 +514,7 @@ def InputData(DirName, CaseName, mTEPES):
     mTEPES.vOutput2ndBlock       = Var(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.nr, within=NonNegativeReals, bounds=lambda mTEPES,sc,p,n,nr: (0.0,mTEPES.pMaxPower2ndBlock[sc,p,n,nr]),                     doc='second block of the unit                         [GW]')
     mTEPES.vReserveUp            = Var(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.nr, within=NonNegativeReals, bounds=lambda mTEPES,sc,p,n,nr: (0.0,mTEPES.pMaxPower2ndBlock[sc,p,n,nr]),                     doc='upward   operating reserve                       [GW]')
     mTEPES.vReserveDown          = Var(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.nr, within=NonNegativeReals, bounds=lambda mTEPES,sc,p,n,nr: (0.0,mTEPES.pMaxPower2ndBlock[sc,p,n,nr]),                     doc='downward operating reserve                       [GW]')
+    mTEPES.vEnergyOutflows       = Var(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.g , within=NonNegativeReals, bounds=lambda mTEPES,sc,p,n,g : (0.0,mTEPES.pMaxPower        [sc,p,n,g ]),                     doc='total outflows of the ESS unit                   [GW]')
     mTEPES.vESSInventory         = Var(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.es, within=NonNegativeReals, bounds=lambda mTEPES,sc,p,n,es: (mTEPES.pMinStorage[sc,p,n,es],mTEPES.pMaxStorage[sc,p,n,es]), doc='ESS inventory                                   [GWh]')
     mTEPES.vESSSpillage          = Var(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.es, within=NonNegativeReals,                                                                                                doc='ESS spillage                                    [GWh]')
     mTEPES.vESSTotalCharge       = Var(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.es, within=NonNegativeReals, bounds=lambda mTEPES,sc,p,n,es: (0.0,mTEPES.pMaxCharge       [es]       ),                     doc='ESS total charge power                           [GW]')
@@ -499,6 +542,15 @@ def InputData(DirName, CaseName, mTEPES):
     else:
         mTEPES.vNetworkInvest    = Var(                               mTEPES.lc, within=Binary,                                                                                                          doc='network    investment decision exists in a year {0,1}')
 
+    if mTEPES.pIndBinLineCommit == 0:
+        mTEPES.vLineCommit       = Var(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.la, within=UnitInterval,                                                                                                    doc='line switching decision exists in a year [0,1]'       )
+        mTEPES.vLineOnState      = Var(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.la, within=UnitInterval,                                                                                                    doc='on state    of the line                         [0,1]')
+        mTEPES.vLineOffState     = Var(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.la, within=UnitInterval,                                                                                                    doc='off state   of the line                         [0,1]')
+    else:
+        mTEPES.vLineCommit       = Var(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.la, within=Binary,                                                                                                          doc='line switching decision exists in a year {0,1}'       )
+        mTEPES.vLineOnState      = Var(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.la, within=Binary,                                                                                                          doc='on state    of the line                         {0,1}')
+        mTEPES.vLineOffState     = Var(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.la, within=Binary,                                                                                                          doc='off state   of the line                         {0,1}')
+
     # relax binary condition in generation and network investment decisions
     for gc in mTEPES.gc:
         if mTEPES.pIndBinGenInvest != 0 and pIndBinUnitInvest[gc] == 0:
@@ -507,12 +559,19 @@ def InputData(DirName, CaseName, mTEPES):
         if mTEPES.pIndBinNetInvest != 0 and pIndBinLineInvest[lc] == 0:
             mTEPES.vNetworkInvest   [lc].domain = UnitInterval
 
+    # by default existing AC and DC lines are always committed
+    for sc,p,n,ni,nf,cc in list(mTEPES.sc*mTEPES.p*mTEPES.n*mTEPES.le):
+        if mTEPES.pIndBinSwitching[ni,nf,cc] == 0:
+            mTEPES.vLineCommit  [sc,p,n,ni,nf,cc].fix(1)
+            mTEPES.vLineOnState [sc,p,n,ni,nf,cc].fix(0)
+            mTEPES.vLineOffState[sc,p,n,ni,nf,cc].fix(0)
+
     mTEPES.vLineLosses           = Var(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.ll, within=NonNegativeReals, bounds=lambda mTEPES,sc,p,n,*ll: (0.0,0.5*mTEPES.pLineLossFactor[ll]*max(mTEPES.pLineNTCBck[ll],mTEPES.pLineNTCFrw[ll])), doc='half line losses                                 [GW]')
     mTEPES.vFlow                 = Var(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.la, within=Reals,            bounds=lambda mTEPES,sc,p,n,*la: (-mTEPES.pLineNTCBck[la],mTEPES.pLineNTCFrw[la]),                                        doc='flow                                             [GW]')
     mTEPES.vTheta                = Var(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.nd, within=Reals,            bounds=lambda mTEPES,sc,p,n, nd: (-mTEPES.pMaxTheta[sc,p,n,nd],mTEPES.pMaxTheta[sc,p,n,nd]),                              doc='voltage angle                                   [rad]')
 
+    # fix the must-run units and their output
     pMustRun = pMustRun.map(idxDict)
-
     mTEPES.pMustRun = Param(mTEPES.gg, within=Any, initialize=pMustRun.to_dict())
 
     for sc,p,n,g  in mTEPES.sc*mTEPES.p*mTEPES.n*mTEPES.g :
@@ -529,9 +588,9 @@ def InputData(DirName, CaseName, mTEPES):
     for sc,p,n,nr in mTEPES.sc*mTEPES.p*mTEPES.n*mTEPES.nr:
         # must run units or units with no minimum power are always committed and must produce at least their minimum output
         if mTEPES.pMustRun[nr] == 1 or (mTEPES.pMinPower[sc,p,n,nr] == 0.0 and mTEPES.pConstantVarCost[nr] == 0.0):
-            mTEPES.vCommitment[sc,p,n,nr].fix(1)
-            mTEPES.vStartUp   [sc,p,n,nr].fix(0)
-            mTEPES.vShutDown  [sc,p,n,nr].fix(0)
+            mTEPES.vCommitment     [sc,p,n,nr].fix(1)
+            mTEPES.vStartUp        [sc,p,n,nr].fix(0)
+            mTEPES.vShutDown       [sc,p,n,nr].fix(0)
         # if min and max power coincide there are neither second block, nor operating reserve
         if mTEPES.pMaxPower2ndBlock[sc,p,n,nr] ==  0.0:
             mTEPES.vOutput2ndBlock [sc,p,n,nr].fix(0.0)
@@ -546,6 +605,79 @@ def InputData(DirName, CaseName, mTEPES):
             mTEPES.vESSReserveUp  [sc,p,n,es].fix(0.0)
             mTEPES.vESSReserveDown[sc,p,n,es].fix(0.0)
 
+    # thermal and RES units ordered by increasing variable operation cost, excluding reactive generating units
+    if len(mTEPES.tq) > 0:
+        mTEPES.go = pLinearOperCost.sort_values().index.drop(list(mTEPES.gq))
+    else:
+        mTEPES.go = pLinearOperCost.sort_values().index
+
+    # initialize generation output, unit commitment and line switching
+    pInitialOutput = pd.DataFrame([[0.0]*len(mTEPES.gg)]*len(mTEPES.sc*mTEPES.p*mTEPES.n), index=pd.MultiIndex.from_tuples(mTEPES.sc*mTEPES.p*mTEPES.n), columns=list(mTEPES.gg))
+    pInitialUC     = pd.DataFrame([[0  ]*len(mTEPES.gg)]*len(mTEPES.sc*mTEPES.p*mTEPES.n), index=pd.MultiIndex.from_tuples(mTEPES.sc*mTEPES.p*mTEPES.n), columns=list(mTEPES.gg))
+    pInitialSwitch = pd.DataFrame([[0  ]*len(mTEPES.ln)]*len(mTEPES.sc*mTEPES.p*mTEPES.n), index=pd.MultiIndex.from_tuples(mTEPES.sc*mTEPES.p*mTEPES.n), columns=list(mTEPES.la))
+
+    mTEPES.pInitialOutput = Param(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.gg, initialize=pInitialOutput.stack().to_dict(), within=NonNegativeReals, doc='unit initial output',     mutable=True)
+    mTEPES.pInitialUC     = Param(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.gg, initialize=pInitialUC.stack().to_dict(),     within=Boolean,          doc='unit initial commitment', mutable=True)
+    mTEPES.pInitialSwitch = Param(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.ln, initialize=pInitialSwitch.stack().to_dict(), within=Boolean,          doc='line initial switching',  mutable=True)
+
+    for sc,p,st in mTEPES.scc*mTEPES.pp*range(1,int(sum(mTEPES.pDuration.values())/mTEPES.pStageDuration+1)):
+        # activate only scenario to formulate
+        mTEPES.del_component(mTEPES.sc)
+        mTEPES.sc = Set(initialize=mTEPES.scc, ordered=True, doc='scenarios'  , filter=lambda mTEPES,scc: scc in  mTEPES.scc and sc == scc and mTEPES.pScenProb[scc] > 0.0)
+        # activate only period to formulate
+        mTEPES.del_component(mTEPES.p )
+        mTEPES.p  = Set(initialize=mTEPES.pp , ordered=True, doc='periods'    , filter=lambda mTEPES,pp : pp  in p  == pp                                                 )
+        # activate only load levels of this stage
+        mTEPES.del_component(mTEPES.n )
+        mTEPES.n  = Set(initialize=mTEPES.nn , ordered=True, doc='load levels', filter=lambda mTEPES,nn : nn  in list(mTEPES.pDuration) and mTEPES.nn.ord(nn) > (st-1)*mTEPES.pStageDuration and mTEPES.nn.ord(nn) <= st*mTEPES.pStageDuration)
+
+        # commit the units and their output at the first load level of each stage
+        pSystemOutput = 0.0
+        n1 = next(iter(mTEPES.sc*mTEPES.p*mTEPES.n))
+        for nr in mTEPES.nr:
+            if pSystemOutput < sum(mTEPES.pDemand[n1,nd] for nd in mTEPES.nd) and mTEPES.pMustRun[nr] == 1:
+                mTEPES.pInitialOutput[n1,nr] = mTEPES.pMaxPower[n1,nr]
+                mTEPES.pInitialUC    [n1,nr] = 1
+                pSystemOutput               += mTEPES.pInitialOutput[n1,nr]()
+
+        # determine the initial committed units and their output at the first load level of each stage
+        for go in mTEPES.go:
+            if pSystemOutput < sum(mTEPES.pDemand[n1,nd] for nd in mTEPES.nd) and mTEPES.pMustRun[go] != 1:
+                if go in mTEPES.r:
+                    mTEPES.pInitialOutput[n1,go] = mTEPES.pMaxPower[n1,go]
+                else:
+                    mTEPES.pInitialOutput[n1,go] = mTEPES.pMinPower[n1,go]
+                mTEPES.pInitialUC[n1,go] = 1
+                pSystemOutput = pSystemOutput + mTEPES.pInitialOutput[n1,go]()
+
+        # determine the initial committed lines
+        for la in mTEPES.la:
+            if la in mTEPES.lc:
+                mTEPES.pInitialSwitch[n1,la] = 0
+            else:
+                mTEPES.pInitialSwitch[n1,la] = 1
+
+        # fixing the ESS inventory at the last load level
+        for sc,p,es in mTEPES.sc*mTEPES.p*mTEPES.es:
+            mTEPES.vESSInventory[sc,p,mTEPES.n.last(),es].fix(pInitialInventory[es])
+
+    # activate all the scenarios, periods and load levels again
+    mTEPES.del_component(mTEPES.sc)
+    mTEPES.del_component(mTEPES.p )
+    mTEPES.del_component(mTEPES.n )
+    mTEPES.sc = Set(initialize=mTEPES.scc, ordered=True, doc='scenarios',   filter=lambda mTEPES,scc: scc in mTEPES.scc and pScenProb[scc] > 0.0)
+    mTEPES.p  = Set(initialize=mTEPES.pp,  ordered=True, doc='periods',     filter=lambda mTEPES,pp : pp  in p == pp                            )
+    mTEPES.n  = Set(initialize=mTEPES.nn,  ordered=True, doc='load levels', filter=lambda mTEPES,nn : nn  in list(mTEPES.pDuration)             )
+
+    # fixing the ESS inventory at the end of the following pCycleTimeStep (daily, weekly, monthly), i.e., for daily ESS is fixed at the end of the week, for weekly/monthly ESS is fixed at the end of the year
+    for sc,p,n,es in mTEPES.sc*mTEPES.p*mTEPES.n*mTEPES.es:
+         if pStorageType[es] == 'Daily'   and mTEPES.n.ord(n) % int( 168/pTimeStep) == 0:
+             mTEPES.vESSInventory[sc,p,n,es].fix(pInitialInventory[es])
+         if pStorageType[es] == 'Weekly'  and mTEPES.n.ord(n) % int(8736/pTimeStep) == 0:
+             mTEPES.vESSInventory[sc,p,n,es].fix(pInitialInventory[es])
+         if pStorageType[es] == 'Monthly' and mTEPES.n.ord(n) % int(8736/pTimeStep) == 0:
+             mTEPES.vESSInventory[sc,p,n,es].fix(pInitialInventory[es])
+
     # if no operating reserve is required no variables are needed
     for sc,p,n,ar,nr in mTEPES.sc*mTEPES.p*mTEPES.n*mTEPES.ar*mTEPES.nr:
         if (ar,nr) in mTEPES.a2g:
@@ -560,46 +692,6 @@ def InputData(DirName, CaseName, mTEPES):
             if mTEPES.pOperReserveDw  [sc,p,n,ar] ==  0.0:
                 mTEPES.vESSReserveDown[sc,p,n,es].fix(0.0)
 
-    # fix the must-run units and their output
-    mTEPES.pInitialOutput = pd.Series([0.0]*len(mTEPES.nr), index=list(mTEPES.nr))
-    mTEPES.pInitialUC     = pd.Series([0.0]*len(mTEPES.nr), index=list(mTEPES.nr))
-    pSystemOutput  = 0.0
-    n1 = next(iter(mTEPES.sc*mTEPES.p*mTEPES.n))
-    for nr in mTEPES.nr:
-        if pSystemOutput < sum(mTEPES.pDemand[n1,nd] for nd in mTEPES.nd) and mTEPES.pMustRun[nr] == 1:
-            mTEPES.pInitialOutput[nr] = mTEPES.pMaxPower[n1,nr]
-            mTEPES.pInitialUC    [nr] = 1
-            pSystemOutput            += mTEPES.pInitialOutput[nr]
-
-    # thermal and variable units ordered by increasing variable operation cost
-    if len(mTEPES.tq) > 0:
-        mTEPES.go = pLinearOperCost.sort_values().index.drop(list(mTEPES.gq))
-    else:
-        mTEPES.go = pLinearOperCost.sort_values().index
-
-    # determine the initial committed units and their output
-    for go in mTEPES.go:
-        if pSystemOutput < sum(mTEPES.pDemand[n1,nd] for nd in mTEPES.nd) and mTEPES.pMustRun[go] != 1:
-            if go in mTEPES.r:
-                mTEPES.pInitialOutput[go] = mTEPES.pMaxPower[n1,go]
-            else:
-                mTEPES.pInitialOutput[go] = mTEPES.pMinPower[n1,go]
-            mTEPES.pInitialUC    [go] = 1
-            pSystemOutput      = pSystemOutput + mTEPES.pInitialOutput[go]
-
-    # fixing the ESS inventory at the last load level
-    for sc,p,es in mTEPES.sc*mTEPES.p*mTEPES.es:
-        mTEPES.vESSInventory[sc,p,mTEPES.n.last(),es].fix(pInitialInventory[es])
-
-    # fixing the ESS inventory at the end of the following pCycleTimeStep (daily, weekly, monthly), i.e., for daily ESS is fixed at the end of the week, for weekly/monthly ESS is fixed at the end of the year
-    for sc,p,n,es in mTEPES.sc*mTEPES.p*mTEPES.n*mTEPES.es:
-         if pStorageType[es] == 'Daily'   and mTEPES.n.ord(n) % int( 168/pTimeStep) == 0:
-             mTEPES.vESSInventory[sc,p,n,es].fix(pInitialInventory[es])
-         if pStorageType[es] == 'Weekly'  and mTEPES.n.ord(n) % int(8736/pTimeStep) == 0:
-             mTEPES.vESSInventory[sc,p,n,es].fix(pInitialInventory[es])
-         if pStorageType[es] == 'Monthly' and mTEPES.n.ord(n) % int(8736/pTimeStep) == 0:
-             mTEPES.vESSInventory[sc,p,n,es].fix(pInitialInventory[es])
-
     # fixing the voltage angle of the reference node for each scenario, period, and load level
     for sc,p,n in mTEPES.sc*mTEPES.p*mTEPES.n:
         mTEPES.vTheta[sc,p,n,mTEPES.rf.first()].fix(0.0)
@@ -611,4 +703,4 @@ def InputData(DirName, CaseName, mTEPES):
 
     SettingUpDataTime = time.time() - StartTime
     StartTime         = time.time()
-    print('Setting up input data                 ... ', round(SettingUpDataTime), 's')
+    print('Setting up input data                  ... ', round(SettingUpDataTime), 's')
