@@ -1,4 +1,4 @@
-""" Open Generation and Transmission Operation and Expansion Planning Model with RES and ESS (openTEPES) - April 7, 2021
+""" Open Generation and Transmission Operation and Expansion Planning Model with RES and ESS (openTEPES) - April 10, 2021
 """
 
 import time
@@ -73,15 +73,15 @@ def GenerationOperationModelFormulation(mTEPES, st):
 
     def eInstalGenCap(mTEPES,sc,p,n,gc):
         if mTEPES.pMaxPower[sc,p,n,gc]:
-            return mTEPES.vTotalOutput   [sc,p,n,gc] / mTEPES.pMaxPower[sc,p,n,gc] <= mTEPES.vGenerationInvest[gc]
+            return mTEPES.vTotalOutput[sc,p,n,gc] / mTEPES.pMaxPower[sc,p,n,gc] <= mTEPES.vGenerationInvest[gc]
         else:
-            return mTEPES.vTotalOutput   [sc,p,n,gc]                               <= 0.0
+            return mTEPES.vTotalOutput[sc,p,n,gc]                               <= 0.0
     setattr(mTEPES, 'eInstalGenCap_stage'+str(st), Constraint(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.gc, rule=eInstalGenCap, doc='output      if installed gen unit [p.u.]'))
 
     print('eInstalGenCap         ... ', len(getattr(mTEPES, 'eInstalGenCap_stage'+str(st))), ' rows')
 
     def eInstalConESS(mTEPES,sc,p,n,ec):
-        return mTEPES.vESSTotalCharge[sc,p,n,ec] / mTEPES.pMaxCharge      [ec] <= mTEPES.vGenerationInvest[ec]
+        return mTEPES.vESSTotalCharge [sc,p,n,ec] / mTEPES.pMaxCharge      [ec] <= mTEPES.vGenerationInvest[ec]
     setattr(mTEPES, 'eInstalConESS_stage'+str(st), Constraint(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.ec, rule=eInstalConESS, doc='consumption if installed ESS unit [p.u.]'))
 
     print('eInstalConESS         ... ', len(getattr(mTEPES, 'eInstalConESS_stage'+str(st))), ' rows')
@@ -373,7 +373,7 @@ def NetworkSwitchingModelFormulation(mTEPES, st):
 
     def eSWOnOff(mTEPES,sc,p,n,ni,nf,cc):
         if   n == mTEPES.n.first() and mTEPES.pIndBinSwitching[ni,nf,cc] == 1:
-            return mTEPES.vLineCommit[sc,p,n,ni,nf,cc] - mTEPES.pInitialSwitch[ni,nf,cc]                    == mTEPES.vLineOnState[sc,p,n,ni,nf,cc] - mTEPES.vLineOffState[sc,p,n,ni,nf,cc]
+            return mTEPES.vLineCommit[sc,p,n,ni,nf,cc] - mTEPES.pInitialSwitch[sc,p,n,ni,nf,cc]             == mTEPES.vLineOnState[sc,p,n,ni,nf,cc] - mTEPES.vLineOffState[sc,p,n,ni,nf,cc]
         elif n != mTEPES.n.first() and mTEPES.pIndBinSwitching[ni,nf,cc] == 1:
             return mTEPES.vLineCommit[sc,p,n,ni,nf,cc] - mTEPES.vLineCommit[sc,p,mTEPES.n.prev(n),ni,nf,cc] == mTEPES.vLineOnState[sc,p,n,ni,nf,cc] - mTEPES.vLineOffState[sc,p,n,ni,nf,cc]
         else:
